defmodule SqlMembershipProvider.Membership do
  @moduledoc """
  Struct for representing a user's membership to an application and their authentication information.
  """

  use Ecto.Schema
  import Ecto.Query, only: [from: 2]

  @primary_key {:user_id, :binary_id, []}
  @field_source_mapper fn name -> name |> Atom.to_string() |> Macro.camelize() end

  schema "aspnet_Membership" do
    field(:application_id, :binary_id)
    field(:password, :string)
    field(:password_format, :integer)
    field(:password_salt, :string)
    field(:mobile_pin, :string)
    field(:email, :string)
    field(:lowered_email, :string)
    field(:password_question, :string)
    field(:password_answer, :string)
    field(:is_approved, :boolean)
    field(:is_locked_out, :boolean)
    field(:create_date, :utc_datetime)
    field(:last_login_date, :utc_datetime)
    field(:last_password_changed_date, :utc_datetime)
    field(:last_lockout_date, :utc_datetime)
    field(:failed_password_attempt_count, :integer)
    field(:failed_password_attempt_window_start, :utc_datetime)
    field(:failed_password_answer_attempt_count, :integer)
    field(:failed_password_answer_attempt_window_start, :utc_datetime)
    field(:comment, :string)

    belongs_to(:user, SqlMembershipProvider.User,
      define_field: false,
      primary_key: true,
      references: :user_id
    )
  end

  @type t() :: %SqlMembershipProvider.Membership{
          application_id: String.t(),
          user_id: String.t(),
          password: String.t(),
          password_format: integer(),
          password_salt: String.t(),
          mobile_pin: String.t() | nil,
          email: String.t() | nil,
          lowered_email: String.t() | nil,
          password_question: String.t() | nil,
          password_answer: String.t() | nil,
          is_approved: boolean(),
          is_locked_out: boolean(),
          create_date: tuple(),
          last_login_date: tuple(),
          last_password_changed_date: tuple(),
          last_lockout_date: tuple(),
          failed_password_attempt_count: integer(),
          failed_password_attempt_window_start: tuple(),
          failed_password_answer_attempt_count: integer(),
          failed_password_answer_attempt_window_start: tuple(),
          comment: String.t() | nil
        }

  @doc """
  Fetch a membership by user id.
  """
  @spec find_by_user_id(String.t()) :: Ecto.Query.t()
  def find_by_user_id(user_id) when is_binary(user_id) do
    from(
      m in SqlMembershipProvider.Membership,
      where: m.user_id == ^user_id
    )
  end

  @doc """
  Check if a plaintext password matches a user's hashed password.
  """
  @spec is_password_valid?(
          SqlMembershipProvider.Membership.t(),
          String.t()
        ) :: boolean()
  def is_password_valid?(
        %SqlMembershipProvider.Membership{
          password: hashed_password,
          password_format: format,
          password_salt: salt
        },
        password
      )
      when is_binary(password) do
    hash_password(password, format, salt) == hashed_password
  end

  defp hash_password(password, 1, salt) do
    salt = Base.decode64!(salt)

    password =
      String.to_charlist(password)
      |> :unicode.characters_to_binary(:utf8, {:utf16, :little})

    :crypto.hash(:sha, salt <> password)
    |> Base.encode64()
  end

  defp hash_password(_password, format, _salt),
    do: raise(SqlMembershipProvider.UnsupportedPasswordFormatError, format)
end
